/* gqlist.c 
 * Steven Rostedt.
 *
 * Since this is a wrapper for glist, and glist is GNU: 
 * I'm showing a GNU statement here!
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <stdio.h>
#include "glib.h"
#include "gqlist.h"

static GList *
qlist_get_index (GQList *gqlist, gint indx)
{
  gint f,c,l,count,inc;
  GList *glist;
  enum {QLIST_FIRST, QLIST_CURR, QLIST_LAST} use;

  if (!gqlist || (indx < 0) || (indx >= gqlist->size)) 
    return NULL;

  /*
   * find the closest way to find the index.
   */
  f = indx;
  l = (gqlist->size-1) - indx;
  c = (gqlist->index < 0) ? -1 :
    (gqlist->index > indx) ? gqlist->index - indx:
    indx - gqlist->index;

  use = (c < 0) || (c > l) ? 
    ((l > f) ? QLIST_FIRST : QLIST_LAST) :
    ((c > f) ? QLIST_FIRST : QLIST_CURR);
  
  switch (use) {
  case QLIST_FIRST:
    glist = gqlist->first;
    count = f;
    inc = TRUE;
    break;

  case QLIST_LAST:
    glist = gqlist->last;
    count = l;
    inc = FALSE;
    break;

  case QLIST_CURR:
    glist = gqlist->current;
    count = c;
    inc = (gqlist->index > indx) ? FALSE : TRUE;
    break;

  } /* endswitch */


  while ((glist) && (count > 0)) {
    if (inc) glist = glist->next;
    else glist = glist->prev;
    count--;
  } /* endwhile */

  gqlist->current = glist;
  gqlist->index = indx;

  return glist;
} /* end qlist_get_index */

static void
qlist_sort_me (GList *left, GList *right, 
	       GCompareFunc func)
{
  GList *last,*i;
  gpointer swap;

  if (!left || !right || (left == right)) return;

  last = left;

  for (i=left->next; (i) && (i!=right->next); i=i->next)
    if ((i!=last) && (*func)(i->data,left->data) < 0) {
      last = last->next;
      swap = last->data;
      last->data = i->data;
      i->data = swap;
    } /* end if */

  if (left!=last) {
    swap = left->data;
    left->data = last->data;
    last->data = swap;
  } /* endif */

  if (last!=left)
    qlist_sort_me(left, last->prev,func);
  if (last!=right)
    qlist_sort_me(last->next,right,func);
  
} /* end qlist_sort_me */

GQList *
g_qlist_new(GList* glist)
{
  GQList *gqlist;

  gqlist = g_new(GQList,1);

  if (glist) {
    gqlist->first = glist;
    gqlist->size = 0;
    while(glist) {
      gqlist->last = glist;
      gqlist->size++;
      glist = glist->next;
    } /* endwhile */
  } else {
    gqlist->first = gqlist->last = NULL;
    gqlist->size = 0;
  } /* endif */

  gqlist->current = NULL;
  gqlist->index = -1;

  return gqlist;
} /* end g_qlist_new */

void 
g_qlist_free (GQList *gqlist)
{
  g_list_free(gqlist->first);

  g_free(gqlist);
} /* end g_qlist_free */

GQList *
g_qlist_append (GQList *gqlist, gpointer data)
{
  if (!gqlist) {
    gqlist = g_qlist_new(g_list_append(NULL,data));
    return gqlist;
  } /* endif */

  if (gqlist->last) {
    /* 
     * We have content.
     */
    g_list_append(gqlist->last,data);
    gqlist->last = gqlist->last->next;
  } else {
    gqlist->first = 
      gqlist->last = g_list_append(NULL,data);
  } /* endif */
  gqlist->size++;

  return gqlist;
} /* end g_qlist_append */

GQList *
g_qlist_prepend (GQList *gqlist, gpointer data)
{
  if (!gqlist) {
    gqlist = g_qlist_new(g_list_append(NULL,data));
    return gqlist;
  } /* endif */

  if (gqlist->first) {
    /* 
     * We have content.
     */
    g_list_prepend(gqlist->first,data);
    gqlist->first = gqlist->first->prev;
  } else {
    gqlist->first = 
      gqlist->last = g_list_prepend(NULL,data);
  } /* endif */
  gqlist->size++;

  return gqlist;
} /* end g_qlist_prepend */

GQList *
g_qlist_insert_index (GQList *gqlist, gint indx, gpointer data)
{

  if (!gqlist) {
    /* print some error here! */
    fprintf (stderr,"ouch! GQList NOT INITIALIZED! (g_qlist_insert_index)\n");
    return NULL;
  } /* endif */

  if (indx == 0) {
    /* 
     * same as prepend.
     */
    gqlist = g_qlist_prepend (gqlist,data);
  } else if ((indx == -1) || (indx == (gqlist->size-1))) {
    /*
     * same as append.
     */
    gqlist = g_qlist_append (gqlist,data);
  } else {
    GList *glist;
    /* find the position. */
    glist = qlist_get_index (gqlist,indx);
    if (!glist) {
      /*
       * bad index.
       */
      return NULL;
    } /* endif */
    g_list_prepend(glist,data);
    /* 
     * Don't worry about checking first or last
     * we already did that with the previous logic
     */
  } /* endif */

  gqlist->size++;

  return gqlist;
} /* end g_qlist_insert_index */  

gpointer
g_qlist_remove_first (GQList *gqlist)
{
  gpointer data;
  GList *glist;

  if (!gqlist || !(gqlist->first)) return NULL;

  glist = gqlist->first;
  data = glist->data;
  gqlist->first = glist->next;
  glist->next = glist->prev = NULL;
  g_list_free(glist);

  if (gqlist->first == NULL) 
    gqlist->last = NULL;

  /* reset indexing */
  gqlist->current = NULL;
  gqlist->index = -1;

  gqlist->size--;

  return data;
} /* end g_list_remove_first */

gpointer
g_qlist_remove_last (GQList *gqlist)
{
  gpointer data;
  GList *glist;

  if (!gqlist || !(gqlist->last)) return NULL;

  glist = gqlist->last;
  data = glist->data;
  gqlist->last = glist->prev;
  glist->next = glist->prev = NULL;
  g_list_free(glist);

  if (gqlist->last == NULL) 
    gqlist->first = NULL;

  /* reset indexing */
  gqlist->current = NULL;
  gqlist->index = -1;

  gqlist->size--;

  return data;
} /* end g_list_remove_first */

gpointer
g_qlist_remove_index (GQList *gqlist, gint indx)
{
  GList *glist;
  gpointer data;

  if (!gqlist || !(gqlist->first)) return NULL;

  if (indx == 0) {
    /*
     * Same as g_qlist_remove_first 
     */
    return g_qlist_remove_first(gqlist);
  }
  if ((indx == -1) || (indx == (gqlist->size-1))) {
    /*
     * Same as g_qlist_remove_last
     */
    return g_qlist_remove_last(gqlist);
  }
  /* 
   * All the rest are within the first and last.
   */
  glist = qlist_get_index(gqlist,indx);

  if (!glist) {
    /* bad index */
    return NULL;
  } /* end */

  data = glist->data;
  glist->prev->next = glist->next;
  glist->next->prev = glist->prev;
  glist->next = glist->prev = NULL;
  g_list_free (glist);

  /* reset indexing */
  gqlist->current = NULL;
  gqlist->index = -1;

  gqlist->size--;

  return data;
} /* end g_qlist_remove_index */

gpointer 
g_qlist_get_first (GQList *gqlist)
{
  if (!(gqlist) || !(gqlist->first)) return NULL;

  return (gqlist->first->data);
} /* end g_qlist_get_first */

gpointer 
g_qlist_get_last (GQList *gqlist)
{
  if (!(gqlist) || !(gqlist->last)) return NULL;

  return (gqlist->last->data);
} /* end g_qlist_get_last */

gpointer 
g_qlist_get_data (GQList *gqlist, gint indx)
{
  GList *glist;
  if (!(gqlist) || !(gqlist->first)) return NULL;

  glist = qlist_get_index(gqlist,indx);

  if (!glist) return NULL;

  return (glist->data);
} /* end g_qlist_get_data */


gint
g_qlist_size (GQList *gqlist)
{ 
  /*
   * I just don't like accessing the objects directly
   * from the structure. 
   * I like C++ type of C programming :)
   */
  if (!gqlist) return 0;
  return gqlist->size;
} /* end g_qlist_size */


void
g_qlist_sort (GQList *gqlist,
	      GCompareFunc func)
{
  qlist_sort_me(gqlist->first, gqlist->last,func);
} /* endif */

